# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .inferrer import IncrDetectionInferrer
from .trainer import IncrDetectionTrainer

__all__ = ["IncrDetectionTrainer", "IncrDetectionInferrer"]
