# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmov_rpn_head import MMOVRPNHead
from .mmov_ssd_head import MMOVSSDHead

# flake8: noqa
from .mmov_yolov3_head import MMOVYOLOV3Head
