"""OpenVINO Training Extensions."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

__version__ = "1.0.0"
# NOTE: Sync w/ otx/api/usecases/exportable_code/demo/requirements.txt
