# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .cls_head import ClsHead
from .conv_head import ConvClsHead

# flake8: noqa
from .mmov_cls_head import MMOVClsHead
