# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .inferrer import IncrClsInferrer
from .trainer import IncrClsTrainer

__all__ = ["IncrClsTrainer", "IncrClsInferrer"]
