"""Entities containing shapes."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# from .rectangle import *
# from .circle import *
# from .polygon import *
# from .shape import *
