"""
Requirements module.
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


class Requirements:
    """
    Dummy requirement.
    """

    REQ_1 = "Dummy requirement"
