"""OTX Adapters for action recognition detectors - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from .fast_rcnn import AVAFastRCNN

__all__ = ["AVAFastRCNN"]
