"""Base configuration parameters for Draem."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from .configuration import DraemAnomalyBaseConfig

__all__ = ["DraemAnomalyBaseConfig"]
