"""Task Initialization of OTX Classification."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import otx.algorithms.classification.adapters.mmcls as MPAData
import otx.algorithms.common.adapters.mmcv.models as OTXBackbones
import otx.mpa.cls as MPAClassification

from .inference import ClassificationInferenceTask
from .nncf import ClassificationNNCFTask
from .openvino import ClassificationOpenVINOTask
from .train import ClassificationTrainTask

__all__ = [
    "MPAClassification",
    "OTXBackbones",
    "MPAData",
    "ClassificationOpenVINOTask",
    "ClassificationInferenceTask",
    "ClassificationTrainTask",
    "ClassificationNNCFTask",
]
