"""Utilities and use cases built on top of OTX API.

.. automodule:: otx.api.usecases.adapters
   :members:
   :undoc-members:

"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
