"""OTX Algorithms - Utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .labels_utils import get_multihead_class_info

__all__ = ["get_multihead_class_info"]
